@@ -25,11 +25,9 @@ import GHC.CmmToLlvm.Mangler
 import GHC.StgToCmm.CgUtils ( fixStgRegisters )
 import GHC.Cmm
 import GHC.Cmm.Dataflow.Collections
-import GHC.Cmm.Ppr
 
 import GHC.Utils.BufHandle
 import GHC.Driver.Session
-import GHC.Platform ( platformArch, Arch(..) )
 import GHC.Utils.Error
 import GHC.Data.FastString
 import GHC.Utils.Outputable
@@ -38,7 +36,7 @@ import GHC.Utils.Logger
 import qualified GHC.Data.Stream as Stream
 
 import Control.Monad ( when, forM_ )
-import Data.Maybe ( fromMaybe, catMaybes )
+import Data.Maybe ( fromMaybe, catMaybes, isNothing )
 import System.IO
 
 -- -----------------------------------------------------------------------------
@@ -68,11 +66,13 @@ llvmCodeGen logger cfg h cmm_stream
            "up to" <+> text (llvmVersionStr supportedLlvmVersionUpperBound) <+> "(non inclusive) is supported." <+>
            "System LLVM version: " <> text (llvmVersionStr ver) $$
            "We will try though..."
-         let isS390X = platformArch (llvmCgPlatform cfg)  == ArchS390X
-         let major_ver = head . llvmVersionList $ ver
-         when (isS390X && major_ver < 10 && doWarn) $ putMsg logger $
-           "Warning: For s390x the GHC calling convention is only supported since LLVM version 10." <+>
-           "You are using LLVM version: " <> text (llvmVersionStr ver)
+
+       when (isNothing mb_ver) $ do
+         let doWarn = llvmCgDoWarn cfg
+         when doWarn $ putMsg logger $
+           "Failed to detect LLVM version!" $$
+           "Make sure LLVM is installed correctly." $$
+           "We will try though..."
 
        -- HACK: the Nothing case here is potentially wrong here but we
        -- currently don't use the LLVM version to guide code generation
@@ -110,8 +110,8 @@ llvmCodeGen' cfg cmm_stream
     header =
       let target  = llvmCgLlvmTarget cfg
           llvmCfg = llvmCgLlvmConfig cfg
-      in     text ("target datalayout = \"" ++ getDataLayout llvmCfg target ++ "\"")
-         $+$ text ("target triple = \"" ++ target ++ "\"")
+      in     (text "target datalayout = \"" <> text (getDataLayout llvmCfg target) <> text "\"")
+         $+$ (text "target triple = \"" <> text target <> text "\"")
 
     getDataLayout :: LlvmConfig -> String -> String
     getDataLayout config target =
