@@ -41,6 +41,7 @@ import GHC.Exts.Heap.InfoTableProf ()
 import GHC.Exts.Heap.ProfInfo.Types
 
 import Data.Bits
+import Data.Foldable (toList)
 import Data.Int
 import Data.Word
 import GHC.Exts
@@ -98,7 +99,7 @@ areBoxesEqual (Box a) (Box b) = case reallyUnsafePtrEqualityUpToTag# a b of
 type Closure = GenClosure Box
 
 -- | This is the representation of a Haskell value on the heap. It reflects
--- <https://gitlab.haskell.org/ghc/ghc/blob/master/includes/rts/storage/Closures.h>
+-- <https://gitlab.haskell.org/ghc/ghc/blob/master/rts/include/rts/storage/Closures.h>
 --
 -- The data type is parametrized by `b`: the type to store references in.
 -- Usually this is a 'Box' with the type synonym 'Closure'.
@@ -228,8 +229,8 @@ data GenClosure b
         , mccPayload :: ![b]            -- ^ Array payload
         }
 
-    -- | An @MVar#@, with a queue of thread state objects blocking on them
-    | MVarClosure
+  -- | An @MVar#@, with a queue of thread state objects blocking on them
+  | MVarClosure
     { info       :: !StgInfoTable
     , queueHead  :: !b              -- ^ Pointer to head of queue
     , queueTail  :: !b              -- ^ Pointer to tail of queue
@@ -265,7 +266,7 @@ data GenClosure b
         , key         :: !b
         , value       :: !b
         , finalizer   :: !b
-        , link        :: !b -- ^ next weak pointer for the capability, can be NULL.
+        , weakLink    :: !(Maybe b) -- ^ next weak pointer for the capability
         }
 
   -- | Representation of StgTSO: A Thread State Object. The values for
@@ -279,6 +280,7 @@ data GenClosure b
       , trec                :: !b
       , blocked_exceptions  :: !b
       , bq                  :: !b
+      , thread_label        :: !(Maybe b)
       -- values
       , what_next           :: !WhatNext
       , why_blocked         :: !WhyBlocked
@@ -386,9 +388,6 @@ data WhyBlocked
   | BlockedOnCCall
   | BlockedOnCCall_Interruptible
   | BlockedOnMsgThrowTo
-#if __GLASGOW_HASKELL__ >= 811 && __GLASGOW_HASKELL__ < 902
-  | BlockedOnIOCompletion
-#endif
   | ThreadMigrating
   | WhyBlockedUnknownValue Word16 -- ^ Please report this as a bug
   deriving (Eq, Show, Generic, Ord)
@@ -423,7 +422,7 @@ allClosures (MVarClosure {..}) = [queueHead,queueTail,value]
 allClosures (IOPortClosure {..}) = [queueHead,queueTail,value]
 allClosures (FunClosure {..}) = ptrArgs
 allClosures (BlockingQueueClosure {..}) = [link, blackHole, owner, queue]
-allClosures (WeakClosure {..}) = [cfinalizers, key, value, finalizer, link]
+allClosures (WeakClosure {..}) = [cfinalizers, key, value, finalizer] ++ Data.Foldable.toList weakLink
 allClosures (OtherClosure {..}) = hvalues
 allClosures _ = []
 
