@@ -1,16 +1,15 @@
 -- Tests succeed if module compiles
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 
-module Test.Format.Compile
-    (
+module Test.Format.Compile (
     ) where
 
 import Data.Time
 
-newtype WrappedUTCTime =
-    MkWrappedUTCTime UTCTime
+newtype WrappedUTCTime
+    = MkWrappedUTCTime UTCTime
     deriving (FormatTime, ParseTime)
 
-newtype Wrapped t =
-    MkWrapped t
+newtype Wrapped t
+    = MkWrapped t
     deriving (FormatTime, ParseTime)
