@@ -20,7 +20,6 @@
 #include "BlockAlloc.h"
 #include "GC.h"
 #include "Compact.h"
-#include "Schedule.h"
 #include "Apply.h"
 #include "Trace.h"
 #include "Weak.h"
@@ -69,6 +68,10 @@
     pointer.
    ------------------------------------------------------------------------- */
 
+static /* STATIC_INLINE */ P_
+thread_obj (const StgInfoTable *info, P_ p);
+
+
 STATIC_INLINE W_
 UNTAG_PTR(W_ p)
 {
@@ -451,6 +454,13 @@ thread_AP_STACK (StgAP_STACK *ap)
     return (P_)ap + sizeofW(StgAP_STACK) + ap->size;
 }
 
+STATIC_INLINE P_
+thread_continuation(StgContinuation *cont)
+{
+    thread_stack(cont->stack, cont->stack + cont->stack_size);
+    return (P_)cont + continuation_sizeW(cont);
+}
+
 static P_
 thread_TSO (StgTSO *tso)
 {
@@ -470,6 +480,10 @@ thread_TSO (StgTSO *tso)
 
     thread_(&tso->trec);
 
+    if (tso->label != NULL) {
+        thread_((StgClosure **)&tso->label);
+    }
+
     thread_(&tso->stackobj);
     return (P_)tso + sizeofW(StgTSO);
 }
@@ -554,6 +568,13 @@ update_fwd_large( bdescr *bd )
       // nothing to follow
       continue;
 
+    // See Note [Black holes in large objects] in Evac.c for why.
+    case BLACKHOLE:
+      {
+        thread_obj(info, p);
+        continue;
+      }
+
     case MUT_ARR_PTRS_CLEAN:
     case MUT_ARR_PTRS_DIRTY:
     case MUT_ARR_PTRS_FROZEN_CLEAN:
@@ -610,6 +631,10 @@ update_fwd_large( bdescr *bd )
         continue;
     }
 
+    case CONTINUATION:
+        thread_continuation((StgContinuation *)p);
+        continue;
+
     default:
       barf("update_fwd_large: unknown/strange object  %d", (int)(info->type));
     }
@@ -796,6 +821,9 @@ thread_obj (const StgInfoTable *info, P_ p)
         return p + sizeofW(StgTRecChunk);
     }
 
+    case CONTINUATION:
+        return thread_continuation((StgContinuation *)p);
+
     default:
         barf("update_fwd: unknown/strange object  %d", (int)(info->type));
         return NULL;
@@ -963,8 +991,6 @@ compact(StgClosure *static_objects,
     // 1. thread the roots
     markCapabilities((evac_fn)thread_root, NULL);
 
-    markScheduler((evac_fn)thread_root, NULL);
-
     // the weak pointer lists...
     for (W_ g = 0; g < RtsFlags.GcFlags.generations; g++) {
         if (generations[g].weak_ptr_list != NULL) {
@@ -978,8 +1004,8 @@ compact(StgClosure *static_objects,
 
     // mutable lists
     for (W_ g = 1; g < RtsFlags.GcFlags.generations; g++) {
-        for (W_ n = 0; n < n_capabilities; n++) {
-            for (bdescr *bd = capabilities[n]->mut_lists[g];
+        for (W_ n = 0; n < getNumCapabilities(); n++) {
+            for (bdescr *bd = getCapability(n)->mut_lists[g];
                  bd != NULL; bd = bd->link) {
                 for (P_ p = bd->start; p < bd->free; p++) {
                     thread((StgClosure **)p);
@@ -1024,7 +1050,7 @@ compact(StgClosure *static_objects,
         debugTrace(DEBUG_gc, "update_fwd:  %d", g);
 
         update_fwd(gen->blocks);
-        for (W_ n = 0; n < n_capabilities; n++) {
+        for (W_ n = 0; n < getNumCapabilities(); n++) {
             update_fwd(gc_threads[n]->gens[g].todo_bd);
             update_fwd(gc_threads[n]->gens[g].part_list);
         }
